# Grid Discretization Helper Copyright (c) 2016, The Regents of the University
# of California, through Lawrence Berkeley National Laboratory (subject to
# receipt of any required approvals from the U.S. Dept. of Energy).  All rights
# reserved.
#
# If you have questions about your rights to use or distribute this software,
# please contact Berkeley Lab's Innovation & Partnerships Office at:
# IPO@lbl.gov.
#
# NOTICE.  This Software was developed under funding from the U.S. Department
# of Energy and the U.S. Government consequently retains certain rights. As
# such, the U.S. Government has been granted for itself and others acting on
# its behalf a paid-up, non-exclusive, irrevocable, worldwide license in the
# Software to reproduce, distribute copies to the public, prepare derivative
# works, and perform publicly and display publicly, and to permit others to do
# so.

cdef class Unit:
    """Defines objects that describe the generic individual units making up the
    problem domain"""

    def __init__(self,grid,number,index,coordinate):
        """Parameters:
        grid: 'Grid' object to which the new 'Unit' object will belong

        number: unique integer label identifying the new 'Unit' object within
          the containing 'Grid' object

        index: integer coordinate(s) of the grid point indicating the position
          of the new 'Unit' object in the rectangular grid defined in the
          containing 'Grid' object

        coordinate: coordinate(s) of the grid point indicating the position of
          the new 'Unit' object in the coordinate space defined by the
          containing 'Grid' object"""

        self.grid = grid
        self.number = number
        self.index = index
        self.coordinate = coordinate

        # The user must modify these later for volumes subject to boundary conditions, etc.
        self.unknown_field = [ True for field in range(self.grid.field_count) ]

        ## These are placeholders for quantities that are determined elsewhere

        # This is determined in Grid.count_neighbors
        self.neighbor_count = 0

        # These are determined in Grid.determine_interior_neighbors
        self.interior_neighbor_local_index = None

        # These are determined in Grid.finalize_degrees_of_freedom
        self.scalar_count = 0
        self.equation_count = 0
        self.equation_count_array = None
        self.DOF_count = 0
        self.equation_to_residual_mapping = None

    ## These are placeholders for routines that the user may supply
            
    cpdef define_variables(self):
        """This method may be overridden in order to define additional
        quantities associated with 'Unit' objects"""
        pass

    cpdef define_interpolants(self):
        """This method may be overridden in order to define operators that
        act on sequences of variable values, with one element for each 'Unit'
        object in the containing 'Grid' object"""
        pass
